/*
 *  linux/drivers/mmc/core/core.c
 *
 *  Copyright (C) 2003-2004 Russell King, All Rights Reserved.
 *  SD support Copyright (C) 2004 Ian Molton, All Rights Reserved.
 *  Copyright (C) 2005-2008 Pierre Ossman, All Rights Reserved.
 *  MMCv4 support Copyright (C) 2006 Philip Langdale, All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
#include <linux/module.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/completion.h>
#include <linux/device.h>
#include <linux/delay.h>
#include <linux/pagemap.h>
#include <linux/err.h>
#include <linux/leds.h>
#include <linux/scatterlist.h>
#include <linux/log2.h>
#include <linux/regulator/consumer.h>

#include <linux/mmc/card.h>
#include <linux/mmc/host.h>
#include <linux/mmc/mmc.h>
#include <linux/mmc/sd.h>

#include "core.h"
#include "bus.h"
#include "host.h"
#include "sdio_bus.h"

#include "mmc_ops.h"
#include "sd_ops.h"
#include "sdio_ops.h"

/*	UHS patch*/
#include "../host/sdhci.h"
#include "JMCR_DEF.h"
#include <linux/pci.h>

static struct workqueue_struct *workqueue;

//#define UHSDBG(fmt, args...)			printk(fmt, ##args)
#define UHSDBG(args...)        
void SetExternOE(struct sdhci_host *sdhost, u8 val);

/*
 * Enabling software CRCs on the data blocks can be a significant (30%)
 * performance cost, and for other reasons may not always be desired.
 * So we allow it it to be disabled.
 */
int use_spi_crc = 1;
module_param(use_spi_crc, bool, 0);

/*
 * We normally treat cards as removed during suspend if they are not
 * known to be on a non-removable bus, to avoid the risk of writing
 * back data to a different card after resume.  Allow this to be
 * overridden if necessary.
 */
#ifdef CONFIG_MMC_UNSAFE_RESUME
int mmc_assume_removable;
#else
int mmc_assume_removable = 1;
#endif
module_param_named(removable, mmc_assume_removable, bool, 0644);
MODULE_PARM_DESC(
	removable,
	"MMC/SD cards are removable and may be removed during suspend");

/*
 * Internal function. Schedule delayed work in the MMC work queue.
 */
static int mmc_schedule_delayed_work(struct delayed_work *work,
				     unsigned long delay)
{
	return queue_delayed_work(workqueue, work, delay);
}

/*
 * Internal function. Flush all scheduled work from the MMC work queue.
 */
static void mmc_flush_scheduled_work(void)
{
	flush_workqueue(workqueue);
}

/**
 *	mmc_request_done - finish processing an MMC request
 *	@host: MMC host which completed request
 *	@mrq: MMC request which request
 *
 *	MMC drivers should call this function when they have completed
 *	their processing of a request.
 */
void mmc_request_done(struct mmc_host *host, struct mmc_request *mrq)
{
	struct mmc_command *cmd = mrq->cmd;
	int err = cmd->error;

	if (err && cmd->retries && mmc_host_is_spi(host)) {
		if (cmd->resp[0] & R1_SPI_ILLEGAL_COMMAND)
			cmd->retries = 0;
	}

	if (err && cmd->retries) {
		pr_debug("%s: req failed (CMD%u): %d, retrying...\n",
			mmc_hostname(host), cmd->opcode, err);

		cmd->retries--;
		cmd->error = 0;
		host->ops->request(host, mrq);
	} else {
		led_trigger_event(host->led, LED_OFF);

		pr_debug("%s: req done (CMD%u): %d: %08x %08x %08x %08x\n",
			mmc_hostname(host), cmd->opcode, err,
			cmd->resp[0], cmd->resp[1],
			cmd->resp[2], cmd->resp[3]);

		if (mrq->data) {
			pr_debug("%s:     %d bytes transferred: %d\n",
				mmc_hostname(host),
				mrq->data->bytes_xfered, mrq->data->error);
		}

		if (mrq->stop) {
			pr_debug("%s:     (CMD%u): %d: %08x %08x %08x %08x\n",
				mmc_hostname(host), mrq->stop->opcode,
				mrq->stop->error,
				mrq->stop->resp[0], mrq->stop->resp[1],
				mrq->stop->resp[2], mrq->stop->resp[3]);
		}

		if (mrq->done)
			mrq->done(mrq);
	}
}

EXPORT_SYMBOL(mmc_request_done);

static void
mmc_start_request(struct mmc_host *host, struct mmc_request *mrq)
{
#ifdef CONFIG_MMC_DEBUG
	unsigned int i, sz;
	struct scatterlist *sg;
#endif

	pr_debug("%s: starting CMD%u arg %08x flags %08x\n",
		 mmc_hostname(host), mrq->cmd->opcode,
		 mrq->cmd->arg, mrq->cmd->flags);

	if (mrq->data) {
		pr_debug("%s:     blksz %d blocks %d flags %08x "
			"tsac %d ms nsac %d\n",
			mmc_hostname(host), mrq->data->blksz,
			mrq->data->blocks, mrq->data->flags,
			mrq->data->timeout_ns / 1000000,
			mrq->data->timeout_clks);
	}

	if (mrq->stop) {
		pr_debug("%s:     CMD%u arg %08x flags %08x\n",
			 mmc_hostname(host), mrq->stop->opcode,
			 mrq->stop->arg, mrq->stop->flags);
	}

	WARN_ON(!host->claimed);

	led_trigger_event(host->led, LED_FULL);

	mrq->cmd->error = 0;
	mrq->cmd->mrq = mrq;
	if (mrq->data) {
		BUG_ON(mrq->data->blksz > host->max_blk_size);
		BUG_ON(mrq->data->blocks > host->max_blk_count);
		BUG_ON(mrq->data->blocks * mrq->data->blksz >
			host->max_req_size);

#ifdef CONFIG_MMC_DEBUG
		sz = 0;
		for_each_sg(mrq->data->sg, sg, mrq->data->sg_len, i)
			sz += sg->length;
		BUG_ON(sz != mrq->data->blocks * mrq->data->blksz);
#endif

		mrq->cmd->data = mrq->data;
		mrq->data->error = 0;
		mrq->data->mrq = mrq;
		if (mrq->stop) {
			mrq->data->stop = mrq->stop;
			mrq->stop->error = 0;
			mrq->stop->mrq = mrq;
		}
	}
	host->ops->request(host, mrq);
}

static void mmc_wait_done(struct mmc_request *mrq)
{
	complete(mrq->done_data);
}

/**
 *	mmc_wait_for_req - start a request and wait for completion
 *	@host: MMC host to start command
 *	@mrq: MMC request to start
 *
 *	Start a new MMC custom command request for a host, and wait
 *	for the command to complete. Does not attempt to parse the
 *	response.
 */
void mmc_wait_for_req(struct mmc_host *host, struct mmc_request *mrq)
{
	DECLARE_COMPLETION_ONSTACK(complete);

	mrq->done_data = &complete;
	mrq->done = mmc_wait_done;

	mmc_start_request(host, mrq);

	wait_for_completion(&complete);
}

EXPORT_SYMBOL(mmc_wait_for_req);

/**
 *	mmc_wait_for_cmd - start a command and wait for completion
 *	@host: MMC host to start command
 *	@cmd: MMC command to start
 *	@retries: maximum number of retries
 *
 *	Start a new MMC command for a host, and wait for the command
 *	to complete.  Return any error that occurred while the command
 *	was executing.  Do not attempt to parse the response.
 */
int mmc_wait_for_cmd(struct mmc_host *host, struct mmc_command *cmd, int retries)
{
	struct mmc_request mrq;

	WARN_ON(!host->claimed);

	memset(&mrq, 0, sizeof(struct mmc_request));

	memset(cmd->resp, 0, sizeof(cmd->resp));
	cmd->retries = retries;

	mrq.cmd = cmd;
	cmd->data = NULL;

	mmc_wait_for_req(host, &mrq);

	return cmd->error;
}

EXPORT_SYMBOL(mmc_wait_for_cmd);

/**
 *	mmc_set_data_timeout - set the timeout for a data command
 *	@data: data phase for command
 *	@card: the MMC card associated with the data transfer
 *
 *	Computes the data timeout parameters according to the
 *	correct algorithm given the card type.
 */
void mmc_set_data_timeout(struct mmc_data *data, const struct mmc_card *card)
{
	unsigned int mult;

	/*
	 * SDIO cards only define an upper 1 s limit on access.
	 */
	if (mmc_card_sdio(card)) {
		data->timeout_ns = 1000000000;
		data->timeout_clks = 0;
		return;
	}

	/*
	 * SD cards use a 100 multiplier rather than 10
	 */
	mult = mmc_card_sd(card) ? 100 : 10;

	/*
	 * Scale up the multiplier (and therefore the timeout) by
	 * the r2w factor for writes.
	 */
	if (data->flags & MMC_DATA_WRITE)
		mult <<= card->csd.r2w_factor;

	data->timeout_ns = card->csd.tacc_ns * mult;
	data->timeout_clks = card->csd.tacc_clks * mult;

	/*
	 * SD cards also have an upper limit on the timeout.
	 */
	if (mmc_card_sd(card)) {
		unsigned int timeout_us, limit_us;

		timeout_us = data->timeout_ns / 1000;
		timeout_us += data->timeout_clks * 1000 /
			(card->host->ios.clock / 1000);

		if (data->flags & MMC_DATA_WRITE)
			/*
			 * The limit is really 250 ms, but that is
			 * insufficient for some crappy cards.
			 */
			limit_us = 300000;
		else
			limit_us = 100000;

		/*
		 * SDHC cards always use these fixed values.
		 */
		if (timeout_us > limit_us || mmc_card_blockaddr(card)) {
			data->timeout_ns = limit_us * 1000;
			data->timeout_clks = 0;
		}
	}
	/*
	 * Some cards need very high timeouts if driven in SPI mode.
	 * The worst observed timeout was 900ms after writing a
	 * continuous stream of data until the internal logic
	 * overflowed.
	 */
	if (mmc_host_is_spi(card->host)) {
		if (data->flags & MMC_DATA_WRITE) {
			if (data->timeout_ns < 1000000000)
				data->timeout_ns = 1000000000;	/* 1s */
		} else {
			if (data->timeout_ns < 100000000)
				data->timeout_ns =  100000000;	/* 100ms */
		}
	}
}
EXPORT_SYMBOL(mmc_set_data_timeout);

/**
 *	mmc_align_data_size - pads a transfer size to a more optimal value
 *	@card: the MMC card associated with the data transfer
 *	@sz: original transfer size
 *
 *	Pads the original data size with a number of extra bytes in
 *	order to avoid controller bugs and/or performance hits
 *	(e.g. some controllers revert to PIO for certain sizes).
 *
 *	Returns the improved size, which might be unmodified.
 *
 *	Note that this function is only relevant when issuing a
 *	single scatter gather entry.
 */
unsigned int mmc_align_data_size(struct mmc_card *card, unsigned int sz)
{
	/*
	 * FIXME: We don't have a system for the controller to tell
	 * the core about its problems yet, so for now we just 32-bit
	 * align the size.
	 */
	sz = ((sz + 3) / 4) * 4;

	return sz;
}
EXPORT_SYMBOL(mmc_align_data_size);

/**
 *	mmc_host_enable - enable a host.
 *	@host: mmc host to enable
 *
 *	Hosts that support power saving can use the 'enable' and 'disable'
 *	methods to exit and enter power saving states. For more information
 *	see comments for struct mmc_host_ops.
 */
int mmc_host_enable(struct mmc_host *host)
{
	if (!(host->caps & MMC_CAP_DISABLE))
		return 0;

	if (host->en_dis_recurs)
		return 0;

	if (host->nesting_cnt++)
		return 0;

	cancel_delayed_work_sync(&host->disable);

	if (host->enabled)
		return 0;

	if (host->ops->enable) {
		int err;

		host->en_dis_recurs = 1;
		err = host->ops->enable(host);
		host->en_dis_recurs = 0;

		if (err) {
			pr_debug("%s: enable error %d\n",
				 mmc_hostname(host), err);
			return err;
		}
	}
	host->enabled = 1;
	return 0;
}
EXPORT_SYMBOL(mmc_host_enable);

static int mmc_host_do_disable(struct mmc_host *host, int lazy)
{
	if (host->ops->disable) {
		int err;

		host->en_dis_recurs = 1;
		err = host->ops->disable(host, lazy);
		host->en_dis_recurs = 0;

		if (err < 0) {
			pr_debug("%s: disable error %d\n",
				 mmc_hostname(host), err);
			return err;
		}
		if (err > 0) {
			unsigned long delay = msecs_to_jiffies(err);

			mmc_schedule_delayed_work(&host->disable, delay);
		}
	}
	host->enabled = 0;
	return 0;
}

/**
 *	mmc_host_disable - disable a host.
 *	@host: mmc host to disable
 *
 *	Hosts that support power saving can use the 'enable' and 'disable'
 *	methods to exit and enter power saving states. For more information
 *	see comments for struct mmc_host_ops.
 */
int mmc_host_disable(struct mmc_host *host)
{
	int err;

	if (!(host->caps & MMC_CAP_DISABLE))
		return 0;

	if (host->en_dis_recurs)
		return 0;

	if (--host->nesting_cnt)
		return 0;

	if (!host->enabled)
		return 0;

	err = mmc_host_do_disable(host, 0);
	return err;
}
EXPORT_SYMBOL(mmc_host_disable);

/**
 *	__mmc_claim_host - exclusively claim a host
 *	@host: mmc host to claim
 *	@abort: whether or not the operation should be aborted
 *
 *	Claim a host for a set of operations.  If @abort is non null and
 *	dereference a non-zero value then this will return prematurely with
 *	that non-zero value without acquiring the lock.  Returns zero
 *	with the lock held otherwise.
 */
int __mmc_claim_host(struct mmc_host *host, atomic_t *abort)
{
	DECLARE_WAITQUEUE(wait, current);
	unsigned long flags;
	int stop;

	might_sleep();

	add_wait_queue(&host->wq, &wait);
	spin_lock_irqsave(&host->lock, flags);
	while (1) {
		set_current_state(TASK_UNINTERRUPTIBLE);
		stop = abort ? atomic_read(abort) : 0;
		if (stop || !host->claimed || host->claimer == current)
			break;
		spin_unlock_irqrestore(&host->lock, flags);
		schedule();
		spin_lock_irqsave(&host->lock, flags);
	}
	set_current_state(TASK_RUNNING);
	if (!stop) {
		host->claimed = 1;
		host->claimer = current;
		host->claim_cnt += 1;
	} else
		wake_up(&host->wq);
	spin_unlock_irqrestore(&host->lock, flags);
	remove_wait_queue(&host->wq, &wait);
	if (!stop)
		mmc_host_enable(host);
	return stop;
}

EXPORT_SYMBOL(__mmc_claim_host);

/**
 *	mmc_try_claim_host - try exclusively to claim a host
 *	@host: mmc host to claim
 *
 *	Returns %1 if the host is claimed, %0 otherwise.
 */
int mmc_try_claim_host(struct mmc_host *host)
{
	int claimed_host = 0;
	unsigned long flags;

	spin_lock_irqsave(&host->lock, flags);
	if (!host->claimed || host->claimer == current) {
		host->claimed = 1;
		host->claimer = current;
		host->claim_cnt += 1;
		claimed_host = 1;
	}
	spin_unlock_irqrestore(&host->lock, flags);
	return claimed_host;
}
EXPORT_SYMBOL(mmc_try_claim_host);

static void mmc_do_release_host(struct mmc_host *host)
{
	unsigned long flags;

	spin_lock_irqsave(&host->lock, flags);
	if (--host->claim_cnt) {
		/* Release for nested claim */
		spin_unlock_irqrestore(&host->lock, flags);
	} else {
		host->claimed = 0;
		host->claimer = NULL;
		spin_unlock_irqrestore(&host->lock, flags);
		wake_up(&host->wq);
	}
}

void mmc_host_deeper_disable(struct work_struct *work)
{
	struct mmc_host *host =
		container_of(work, struct mmc_host, disable.work);

	/* If the host is claimed then we do not want to disable it anymore */
	if (!mmc_try_claim_host(host))
		return;
	mmc_host_do_disable(host, 1);
	mmc_do_release_host(host);
}

/**
 *	mmc_host_lazy_disable - lazily disable a host.
 *	@host: mmc host to disable
 *
 *	Hosts that support power saving can use the 'enable' and 'disable'
 *	methods to exit and enter power saving states. For more information
 *	see comments for struct mmc_host_ops.
 */
int mmc_host_lazy_disable(struct mmc_host *host)
{
	if (!(host->caps & MMC_CAP_DISABLE))
		return 0;

	if (host->en_dis_recurs)
		return 0;

	if (--host->nesting_cnt)
		return 0;

	if (!host->enabled)
		return 0;

	if (host->disable_delay) {
		mmc_schedule_delayed_work(&host->disable,
				msecs_to_jiffies(host->disable_delay));
		return 0;
	} else
		return mmc_host_do_disable(host, 1);
}
EXPORT_SYMBOL(mmc_host_lazy_disable);

/**
 *	mmc_release_host - release a host
 *	@host: mmc host to release
 *
 *	Release a MMC host, allowing others to claim the host
 *	for their operations.
 */
void mmc_release_host(struct mmc_host *host)
{
	WARN_ON(!host->claimed);

	mmc_host_lazy_disable(host);

	mmc_do_release_host(host);
}

EXPORT_SYMBOL(mmc_release_host);

/*
 * Internal function that does the actual ios call to the host driver,
 * optionally printing some debug output.
 */
static inline void mmc_set_ios(struct mmc_host *host)
{
	struct mmc_ios *ios = &host->ios;

	pr_debug("%s: clock %uHz busmode %u powermode %u cs %u Vdd %u "
		"width %u timing %u\n",
		 mmc_hostname(host), ios->clock, ios->bus_mode,
		 ios->power_mode, ios->chip_select, ios->vdd,
		 ios->bus_width, ios->timing);
	
	if (host->card)
		host->card->working_clock = ios->clock;
	host->ops->set_ios(host, ios);
}

/*
 * Control chip select pin on a host.
 */
void mmc_set_chip_select(struct mmc_host *host, int mode)
{
	host->ios.chip_select = mode;
	mmc_set_ios(host);
}

/*
 * Sets the host clock to the highest possible frequency that
 * is below "hz".
 */
void mmc_set_clock(struct mmc_host *host, unsigned int hz)
{
//	WARN_ON(hz < host->f_min);

	if (hz > host->f_max)
		hz = host->f_max;

	host->ios.clock = hz;
	mmc_set_ios(host);
	
	if (hz >= JM_CLOCK_100MHZ)
			SetExternOE(mmc_priv(host) ,1);
	else
			SetExternOE(mmc_priv(host) ,0);
}

/*
 * Change the bus mode (open drain/push-pull) of a host.
 */
void mmc_set_bus_mode(struct mmc_host *host, unsigned int mode)
{
	host->ios.bus_mode = mode;
	mmc_set_ios(host);
}

/*
 * Change data bus width of a host.
 */
void mmc_set_bus_width(struct mmc_host *host, unsigned int width)
{
	host->ios.bus_width = width;
	mmc_set_ios(host);
}

/**
 * mmc_vdd_to_ocrbitnum - Convert a voltage to the OCR bit number
 * @vdd:	voltage (mV)
 * @low_bits:	prefer low bits in boundary cases
 *
 * This function returns the OCR bit number according to the provided @vdd
 * value. If conversion is not possible a negative errno value returned.
 *
 * Depending on the @low_bits flag the function prefers low or high OCR bits
 * on boundary voltages. For example,
 * with @low_bits = true, 3300 mV translates to ilog2(MMC_VDD_32_33);
 * with @low_bits = false, 3300 mV translates to ilog2(MMC_VDD_33_34);
 *
 * Any value in the [1951:1999] range translates to the ilog2(MMC_VDD_20_21).
 */
static int mmc_vdd_to_ocrbitnum(int vdd, bool low_bits)
{
	const int max_bit = ilog2(MMC_VDD_35_36);
	int bit;

	if (vdd < 1650 || vdd > 3600)
		return -EINVAL;

	if (vdd >= 1650 && vdd <= 1950)
		return ilog2(MMC_VDD_165_195);

	if (low_bits)
		vdd -= 1;

	/* Base 2000 mV, step 100 mV, bit's base 8. */
	bit = (vdd - 2000) / 100 + 8;
	if (bit > max_bit)
		return max_bit;
	return bit;
}

/**
 * mmc_vddrange_to_ocrmask - Convert a voltage range to the OCR mask
 * @vdd_min:	minimum voltage value (mV)
 * @vdd_max:	maximum voltage value (mV)
 *
 * This function returns the OCR mask bits according to the provided @vdd_min
 * and @vdd_max values. If conversion is not possible the function returns 0.
 *
 * Notes wrt boundary cases:
 * This function sets the OCR bits for all boundary voltages, for example
 * [3300:3400] range is translated to MMC_VDD_32_33 | MMC_VDD_33_34 |
 * MMC_VDD_34_35 mask.
 */
u32 mmc_vddrange_to_ocrmask(int vdd_min, int vdd_max)
{
	u32 mask = 0;

	if (vdd_max < vdd_min)
		return 0;

	/* Prefer high bits for the boundary vdd_max values. */
	vdd_max = mmc_vdd_to_ocrbitnum(vdd_max, false);
	if (vdd_max < 0)
		return 0;

	/* Prefer low bits for the boundary vdd_min values. */
	vdd_min = mmc_vdd_to_ocrbitnum(vdd_min, true);
	if (vdd_min < 0)
		return 0;

	/* Fill the mask, from max bit to min bit. */
	while (vdd_max >= vdd_min)
		mask |= 1 << vdd_max--;

	return mask;
}
EXPORT_SYMBOL(mmc_vddrange_to_ocrmask);

#ifdef CONFIG_REGULATOR

/**
 * mmc_regulator_get_ocrmask - return mask of supported voltages
 * @supply: regulator to use
 *
 * This returns either a negative errno, or a mask of voltages that
 * can be provided to MMC/SD/SDIO devices using the specified voltage
 * regulator.  This would normally be called before registering the
 * MMC host adapter.
 */
int mmc_regulator_get_ocrmask(struct regulator *supply)
{
	int			result = 0;
	int			count;
	int			i;

	count = regulator_count_voltages(supply);
	if (count < 0)
		return count;

	for (i = 0; i < count; i++) {
		int		vdd_uV;
		int		vdd_mV;

		vdd_uV = regulator_list_voltage(supply, i);
		if (vdd_uV <= 0)
			continue;

		vdd_mV = vdd_uV / 1000;
		result |= mmc_vddrange_to_ocrmask(vdd_mV, vdd_mV);
	}

	return result;
}
EXPORT_SYMBOL(mmc_regulator_get_ocrmask);

/**
 * mmc_regulator_set_ocr - set regulator to match host->ios voltage
 * @vdd_bit: zero for power off, else a bit number (host->ios.vdd)
 * @supply: regulator to use
 *
 * Returns zero on success, else negative errno.
 *
 * MMC host drivers may use this to enable or disable a regulator using
 * a particular supply voltage.  This would normally be called from the
 * set_ios() method.
 */
int mmc_regulator_set_ocr(struct regulator *supply, unsigned short vdd_bit)
{
	int			result = 0;
	int			min_uV, max_uV;
	int			enabled;

	enabled = regulator_is_enabled(supply);
	if (enabled < 0)
		return enabled;

	if (vdd_bit) {
		int		tmp;
		int		voltage;

		/* REVISIT mmc_vddrange_to_ocrmask() may have set some
		 * bits this regulator doesn't quite support ... don't
		 * be too picky, most cards and regulators are OK with
		 * a 0.1V range goof (it's a small error percentage).
		 */
		tmp = vdd_bit - ilog2(MMC_VDD_165_195);
		if (tmp == 0) {
			min_uV = 1650 * 1000;
			max_uV = 1950 * 1000;
		} else {
			min_uV = 1900 * 1000 + tmp * 100 * 1000;
			max_uV = min_uV + 100 * 1000;
		}

		/* avoid needless changes to this voltage; the regulator
		 * might not allow this operation
		 */
		voltage = regulator_get_voltage(supply);
		if (voltage < 0)
			result = voltage;
		else if (voltage < min_uV || voltage > max_uV)
			result = regulator_set_voltage(supply, min_uV, max_uV);
		else
			result = 0;

		if (result == 0 && !enabled)
			result = regulator_enable(supply);
	} else if (enabled) {
		result = regulator_disable(supply);
	}

	return result;
}
EXPORT_SYMBOL(mmc_regulator_set_ocr);

#endif

/*
 * Mask off any voltages we don't support and select
 * the lowest voltage
 */
u32 mmc_select_voltage(struct mmc_host *host, u32 ocr)
{
	int bit;

	ocr &= host->ocr_avail;

	bit = ffs(ocr);
	if (bit) {
		bit -= 1;

		ocr &= 3 << bit;

		host->ios.vdd = bit;
		mmc_set_ios(host);
	} else {
		pr_warning("%s: host doesn't support card's voltages\n",
				mmc_hostname(host));
		ocr = 0;
	}

	return ocr;
}

/*
 * Select timing parameters for host.
 */
void mmc_set_timing(struct mmc_host *host, unsigned int timing)
{
	host->ios.timing = timing;
	mmc_set_ios(host);
}

/*
 * Apply power to the MMC stack.  This is a two-stage process.
 * First, we enable power to the card without the clock running.
 * We then wait a bit for the power to stabilise.  Finally,
 * enable the bus drivers and clock to the card.
 *
 * We must _NOT_ enable the clock prior to power stablising.
 *
 * If a host does all the power sequencing itself, ignore the
 * initial MMC_POWER_UP stage.
 */
void mmc_power_up(struct mmc_host *host)
{
	int bit;

	/* If ocr is set, we use it */
	if (host->ocr)
		bit = ffs(host->ocr) - 1;
	else
		bit = fls(host->ocr_avail) - 1;

	host->ios.vdd = bit;
	if (mmc_host_is_spi(host)) {
		host->ios.chip_select = MMC_CS_HIGH;
		host->ios.bus_mode = MMC_BUSMODE_PUSHPULL;
	} else {
		host->ios.chip_select = MMC_CS_DONTCARE;
		host->ios.bus_mode = MMC_BUSMODE_OPENDRAIN;
	}
	host->ios.power_mode = MMC_POWER_UP;
	host->ios.bus_width = MMC_BUS_WIDTH_1;
	host->ios.timing = MMC_TIMING_LEGACY;
	mmc_set_ios(host);

	/*
	 * This delay should be sufficient to allow the power supply
	 * to reach the minimum voltage.
	 */
	mmc_delay(10);

	if (host->f_min > 400000) {
		pr_warning("%s: Minimum clock frequency too high for "
				"identification mode\n", mmc_hostname(host));
		host->ios.clock = host->f_min;
	} else
		host->ios.clock = 400000;

	host->ios.power_mode = MMC_POWER_ON;
	mmc_set_ios(host);

	/*
	 * This delay must be at least 74 clock sizes, or 1 ms, or the
	 * time required to reach a stable voltage.
	 */
	mmc_delay(10);
}
EXPORT_SYMBOL(mmc_power_up);

void mmc_power_off(struct mmc_host *host)
{
	host->ios.clock = 0;
	host->ios.vdd = 0;
	if (!mmc_host_is_spi(host)) {
		host->ios.bus_mode = MMC_BUSMODE_OPENDRAIN;
		host->ios.chip_select = MMC_CS_DONTCARE;
	}
	host->ios.power_mode = MMC_POWER_OFF;
	host->ios.bus_width = MMC_BUS_WIDTH_1;
	host->ios.timing = MMC_TIMING_LEGACY;
	mmc_set_ios(host);
}
EXPORT_SYMBOL(mmc_power_off);
/*
 * Cleanup when the last reference to the bus operator is dropped.
 */
static void __mmc_release_bus(struct mmc_host *host)
{
	BUG_ON(!host);
	BUG_ON(host->bus_refs);
	BUG_ON(!host->bus_dead);

	host->bus_ops = NULL;
}

/*
 * Increase reference count of bus operator
 */
static inline void mmc_bus_get(struct mmc_host *host)
{
	unsigned long flags;

	spin_lock_irqsave(&host->lock, flags);
	host->bus_refs++;
	spin_unlock_irqrestore(&host->lock, flags);
}

/*
 * Decrease reference count of bus operator and free it if
 * it is the last reference.
 */
static inline void mmc_bus_put(struct mmc_host *host)
{
	unsigned long flags;

	spin_lock_irqsave(&host->lock, flags);
	host->bus_refs--;
	if ((host->bus_refs == 0) && host->bus_ops)
		__mmc_release_bus(host);
	spin_unlock_irqrestore(&host->lock, flags);
}

/*
 * Assign a mmc bus handler to a host. Only one bus handler may control a
 * host at any given time.
 */
void mmc_attach_bus(struct mmc_host *host, const struct mmc_bus_ops *ops)
{
	unsigned long flags;

	BUG_ON(!host);
	BUG_ON(!ops);

	WARN_ON(!host->claimed);

	spin_lock_irqsave(&host->lock, flags);

	BUG_ON(host->bus_ops);
	BUG_ON(host->bus_refs);

	host->bus_ops = ops;
	host->bus_refs = 1;
	host->bus_dead = 0;

	spin_unlock_irqrestore(&host->lock, flags);
}

/*
 * Remove the current bus handler from a host. Assumes that there are
 * no interesting cards left, so the bus is powered down.
 */
void mmc_detach_bus(struct mmc_host *host)
{
	unsigned long flags;

	BUG_ON(!host);

	WARN_ON(!host->claimed);
	WARN_ON(!host->bus_ops);

	spin_lock_irqsave(&host->lock, flags);

	host->bus_dead = 1;

	spin_unlock_irqrestore(&host->lock, flags);

	mmc_power_off(host);

	mmc_bus_put(host);
}

/**
 *	mmc_detect_change - process change of state on a MMC socket
 *	@host: host which changed state.
 *	@delay: optional delay to wait before detection (jiffies)
 *
 *	MMC drivers should call this when they detect a card has been
 *	inserted or removed. The MMC layer will confirm that any
 *	present card is still functional, and initialize any newly
 *	inserted.
 */
void mmc_detect_change(struct mmc_host *host, unsigned long delay)
{
#ifdef CONFIG_MMC_DEBUG
	unsigned long flags;
	spin_lock_irqsave(&host->lock, flags);
	WARN_ON(host->removed);
	spin_unlock_irqrestore(&host->lock, flags);
#endif

	mmc_schedule_delayed_work(&host->detect, delay);
}

EXPORT_SYMBOL(mmc_detect_change);

void mmc_init_erase(struct mmc_card *card)
{
	unsigned int sz;

	if (is_power_of_2(card->erase_size))
		card->erase_shift = ffs(card->erase_size) - 1;
	else
		card->erase_shift = 0;

	/*
	 * It is possible to erase an arbitrarily large area of an SD or MMC
	 * card.  That is not desirable because it can take a long time
	 * (minutes) potentially delaying more important I/O, and also the
	 * timeout calculations become increasingly hugely over-estimated.
	 * Consequently, 'pref_erase' is defined as a guide to limit erases
	 * to that size and alignment.
	 *
	 * For SD cards that define Allocation Unit size, limit erases to one
	 * Allocation Unit at a time.  For MMC cards that define High Capacity
	 * Erase Size, whether it is switched on or not, limit to that size.
	 * Otherwise just have a stab at a good value.  For modern cards it
	 * will end up being 4MiB.  Note that if the value is too small, it
	 * can end up taking longer to erase.
	 */
	if (mmc_card_sd(card) && card->ssr.au) {
		card->pref_erase = card->ssr.au;
		card->erase_shift = ffs(card->ssr.au) - 1;
	} else if (card->ext_csd.hc_erase_size) {
		card->pref_erase = card->ext_csd.hc_erase_size;
	} else {
		sz = (card->csd.capacity << (card->csd.read_blkbits - 9)) >> 11;
		if (sz < 128)
			card->pref_erase = 512 * 1024 / 512;
		else if (sz < 512)
			card->pref_erase = 1024 * 1024 / 512;
		else if (sz < 1024)
			card->pref_erase = 2 * 1024 * 1024 / 512;
		else
			card->pref_erase = 4 * 1024 * 1024 / 512;
		if (card->pref_erase < card->erase_size)
			card->pref_erase = card->erase_size;
		else {
			sz = card->pref_erase % card->erase_size;
			if (sz)
				card->pref_erase += card->erase_size - sz;
		}
	}
}

static void mmc_set_mmc_erase_timeout(struct mmc_card *card,
				      struct mmc_command *cmd,
				      unsigned int arg, unsigned int qty)
{
	unsigned int erase_timeout;

	if (card->ext_csd.erase_group_def & 1) {
		/* High Capacity Erase Group Size uses HC timeouts */
		if (arg == MMC_TRIM_ARG)
			erase_timeout = card->ext_csd.trim_timeout;
		else
			erase_timeout = card->ext_csd.hc_erase_timeout;
	} else {
		/* CSD Erase Group Size uses write timeout */
		unsigned int mult = (10 << card->csd.r2w_factor);
		unsigned int timeout_clks = card->csd.tacc_clks * mult;
		unsigned int timeout_us;

		/* Avoid overflow: e.g. tacc_ns=80000000 mult=1280 */
		if (card->csd.tacc_ns < 1000000)
			timeout_us = (card->csd.tacc_ns * mult) / 1000;
		else
			timeout_us = (card->csd.tacc_ns / 1000) * mult;

		/*
		 * ios.clock is only a target.  The real clock rate might be
		 * less but not that much less, so fudge it by multiplying by 2.
		 */
		timeout_clks <<= 1;
		timeout_us += (timeout_clks * 1000) /
			      (card->host->ios.clock / 1000);

		erase_timeout = timeout_us / 1000;

		/*
		 * Theoretically, the calculation could underflow so round up
		 * to 1ms in that case.
		 */
		if (!erase_timeout)
			erase_timeout = 1;
	}

	/* Multiplier for secure operations */
	if (arg & MMC_SECURE_ARGS) {
		if (arg == MMC_SECURE_ERASE_ARG)
			erase_timeout *= card->ext_csd.sec_erase_mult;
		else
			erase_timeout *= card->ext_csd.sec_trim_mult;
	}

	erase_timeout *= qty;

	/*
	 * Ensure at least a 1 second timeout for SPI as per
	 * 'mmc_set_data_timeout()'
	 */
	if (mmc_host_is_spi(card->host) && erase_timeout < 1000)
		erase_timeout = 1000;

	cmd->erase_timeout = erase_timeout;
}

static void mmc_set_sd_erase_timeout(struct mmc_card *card,
				     struct mmc_command *cmd, unsigned int arg,
				     unsigned int qty)
{
	if (card->ssr.erase_timeout) {
		/* Erase timeout specified in SD Status Register (SSR) */
		cmd->erase_timeout = card->ssr.erase_timeout * qty +
				     card->ssr.erase_offset;
	} else {
		/*
		 * Erase timeout not specified in SD Status Register (SSR) so
		 * use 250ms per write block.
		 */
		cmd->erase_timeout = 250 * qty;
	}

	/* Must not be less than 1 second */
	if (cmd->erase_timeout < 1000)
		cmd->erase_timeout = 1000;
}

static void mmc_set_erase_timeout(struct mmc_card *card,
				  struct mmc_command *cmd, unsigned int arg,
				  unsigned int qty)
{
	if (mmc_card_sd(card))
		mmc_set_sd_erase_timeout(card, cmd, arg, qty);
	else
		mmc_set_mmc_erase_timeout(card, cmd, arg, qty);
}

static int mmc_do_erase(struct mmc_card *card, unsigned int from,
			unsigned int to, unsigned int arg)
{
	struct mmc_command cmd;
	unsigned int qty = 0;
	int err;

	/*
	 * qty is used to calculate the erase timeout which depends on how many
	 * erase groups (or allocation units in SD terminology) are affected.
	 * We count erasing part of an erase group as one erase group.
	 * For SD, the allocation units are always a power of 2.  For MMC, the
	 * erase group size is almost certainly also power of 2, but it does not
	 * seem to insist on that in the JEDEC standard, so we fall back to
	 * division in that case.  SD may not specify an allocation unit size,
	 * in which case the timeout is based on the number of write blocks.
	 *
	 * Note that the timeout for secure trim 2 will only be correct if the
	 * number of erase groups specified is the same as the total of all
	 * preceding secure trim 1 commands.  Since the power may have been
	 * lost since the secure trim 1 commands occurred, it is generally
	 * impossible to calculate the secure trim 2 timeout correctly.
	 */
	if (card->erase_shift)
		qty += ((to >> card->erase_shift) -
			(from >> card->erase_shift)) + 1;
	else if (mmc_card_sd(card))
		qty += to - from + 1;
	else
		qty += ((to / card->erase_size) -
			(from / card->erase_size)) + 1;

	if (!mmc_card_blockaddr(card)) {
		from <<= 9;
		to <<= 9;
	}

	memset(&cmd, 0, sizeof(struct mmc_command));
	if (mmc_card_sd(card))
		cmd.opcode = SD_ERASE_WR_BLK_START;
	else
		cmd.opcode = MMC_ERASE_GROUP_START;
	cmd.arg = from;
	cmd.flags = MMC_RSP_SPI_R1 | MMC_RSP_R1 | MMC_CMD_AC;
	err = mmc_wait_for_cmd(card->host, &cmd, 0);
	if (err) {
		printk(KERN_ERR "mmc_erase: group start error %d, "
		       "status %#x\n", err, cmd.resp[0]);
		err = -EINVAL;
		goto out;
	}

	memset(&cmd, 0, sizeof(struct mmc_command));
	if (mmc_card_sd(card))
		cmd.opcode = SD_ERASE_WR_BLK_END;
	else
		cmd.opcode = MMC_ERASE_GROUP_END;
	cmd.arg = to;
	cmd.flags = MMC_RSP_SPI_R1 | MMC_RSP_R1 | MMC_CMD_AC;
	err = mmc_wait_for_cmd(card->host, &cmd, 0);
	if (err) {
		printk(KERN_ERR "mmc_erase: group end error %d, status %#x\n",
		       err, cmd.resp[0]);
		err = -EINVAL;
		goto out;
	}

	memset(&cmd, 0, sizeof(struct mmc_command));
	cmd.opcode = MMC_ERASE;
	cmd.arg = arg;
	cmd.flags = MMC_RSP_SPI_R1B | MMC_RSP_R1B | MMC_CMD_AC;
	mmc_set_erase_timeout(card, &cmd, arg, qty);
	err = mmc_wait_for_cmd(card->host, &cmd, 0);
	if (err) {
		printk(KERN_ERR "mmc_erase: erase error %d, status %#x\n",
		       err, cmd.resp[0]);
		err = -EIO;
		goto out;
	}

	if (mmc_host_is_spi(card->host))
		goto out;

	do {
		memset(&cmd, 0, sizeof(struct mmc_command));
		cmd.opcode = MMC_SEND_STATUS;
		cmd.arg = card->rca << 16;
		cmd.flags = MMC_RSP_R1 | MMC_CMD_AC;
		/* Do not retry else we can't see errors */
		err = mmc_wait_for_cmd(card->host, &cmd, 0);
		if (err || (cmd.resp[0] & 0xFDF92000)) {
			printk(KERN_ERR "error %d requesting status %#x\n",
				err, cmd.resp[0]);
			err = -EIO;
			goto out;
		}
	} while (!(cmd.resp[0] & R1_READY_FOR_DATA) ||
		 R1_CURRENT_STATE(cmd.resp[0]) == 7);
out:
	return err;
}

/**
 * mmc_erase - erase sectors.
 * @card: card to erase
 * @from: first sector to erase
 * @nr: number of sectors to erase
 * @arg: erase command argument (SD supports only %MMC_ERASE_ARG)
 *
 * Caller must claim host before calling this function.
 */
int mmc_erase(struct mmc_card *card, unsigned int from, unsigned int nr,
	      unsigned int arg)
{
	unsigned int rem, to = from + nr;

	if (!(card->host->caps & MMC_CAP_ERASE) ||
	    !(card->csd.cmdclass & CCC_ERASE))
		return -EOPNOTSUPP;

	if (!card->erase_size)
		return -EOPNOTSUPP;

	if (mmc_card_sd(card) && arg != MMC_ERASE_ARG)
		return -EOPNOTSUPP;

	if ((arg & MMC_SECURE_ARGS) &&
	    !(card->ext_csd.sec_feature_support & EXT_CSD_SEC_ER_EN))
		return -EOPNOTSUPP;

	if ((arg & MMC_TRIM_ARGS) &&
	    !(card->ext_csd.sec_feature_support & EXT_CSD_SEC_GB_CL_EN))
		return -EOPNOTSUPP;

	if (arg == MMC_SECURE_ERASE_ARG) {
		if (from % card->erase_size || nr % card->erase_size)
			return -EINVAL;
	}

	if (arg == MMC_ERASE_ARG) {
		rem = from % card->erase_size;
		if (rem) {
			rem = card->erase_size - rem;
			from += rem;
			if (nr > rem)
				nr -= rem;
			else
				return 0;
		}
		rem = nr % card->erase_size;
		if (rem)
			nr -= rem;
	}

	if (nr == 0)
		return 0;

	to = from + nr;

	if (to <= from)
		return -EINVAL;

	/* 'from' and 'to' are inclusive */
	to -= 1;

	return mmc_do_erase(card, from, to, arg);
}
EXPORT_SYMBOL(mmc_erase);

int mmc_can_erase(struct mmc_card *card)
{
	if ((card->host->caps & MMC_CAP_ERASE) &&
	    (card->csd.cmdclass & CCC_ERASE) && card->erase_size)
		return 1;
	return 0;
}
EXPORT_SYMBOL(mmc_can_erase);

int mmc_can_trim(struct mmc_card *card)
{
	if (card->ext_csd.sec_feature_support & EXT_CSD_SEC_GB_CL_EN)
		return 1;
	return 0;
}
EXPORT_SYMBOL(mmc_can_trim);

int mmc_can_secure_erase_trim(struct mmc_card *card)
{
	if (card->ext_csd.sec_feature_support & EXT_CSD_SEC_ER_EN)
		return 1;
	return 0;
}
EXPORT_SYMBOL(mmc_can_secure_erase_trim);

int mmc_erase_group_aligned(struct mmc_card *card, unsigned int from,
			    unsigned int nr)
{
	if (!card->erase_size)
		return 0;
	if (from % card->erase_size || nr % card->erase_size)
		return 0;
	return 1;
}
EXPORT_SYMBOL(mmc_erase_group_aligned);
/*=========================================================================*/
#define OFFSET_DATA_ERROR					-1000
#define OFFSET_CMD_ERROR						-2000
#define ERROR_NONE									0
#define ERROR_RESET_HOST						1
#define ERROR_STOP_ADJUSTMENT				2
#define ERROR_INVAILD_PARAMETER			3
#define ERROR_INVAILD_SETTING					4
#define ERROR_XFERSIZE_MISMACH				5

#define ERROR_READFAIL							10
#define ERROR_DATA_MISMATCH					11

#define ERROR_WRITEFAIL							20

#define ERROR_SETCLOCKFAIL					30
#define ERROR_CARDREMOVE					31
#define ERROR_NO_CLOCK_DELAY_SETTING		32
#define ERROR_NO_SLOT						33
#define ERROR_INVAILD_CARD					34

#define ERROR_NO_MEMORY						40
#define ERROR_MMC_COMMAND_TIMEOUT			41

#define ERROR_UNKNOWN						100

	


#define SD_STATUS_CURRENT_STATE_MASK		0x1E00					/*	remain [12:9]	*/



#define FUNRT_MEMCMP_SAME					0
#define JM_BLOCK_SIZE							512
#define JM_NUM_BLK_MULTI_READ		32



struct sdhci_pci_chip;
struct sdhci_pci_slot;

struct sdhci_pci_fixes {
	unsigned int		quirks;

	int			(*probe)(struct sdhci_pci_chip*);

	int			(*probe_slot)(struct sdhci_pci_slot*);
	void			(*remove_slot)(struct sdhci_pci_slot*, int);

	int			(*suspend)(struct sdhci_pci_chip*,
					pm_message_t);
	int			(*resume)(struct sdhci_pci_chip*);
};

u16 JmReadPciCnfg(struct pci_dev *pdev ,u32 *Value,const u32 OfstJmPCICfg,const u8 length)
{
	u8  val_8;
	u16 val_16;
	u32 val_32;

	val_8 = 0; val_16 = 0; val_32 = 0;
	if (length>4)
		return ERROR_INVAILD_PARAMETER;

	switch (length){
		case 1:
			pci_read_config_byte(pdev, OfstJmPCICfg, &val_8);
			*Value = val_8 ;
			break;
		case 2:
			pci_read_config_word(pdev, OfstJmPCICfg, &val_16);
			*Value = val_16;
			break;
		case 3:
			pci_read_config_dword(pdev, OfstJmPCICfg, &val_32);
			*Value = val_32>>8;					/* or val_32<<8  ????*/
			break;
		case 4:
			pci_read_config_dword(pdev, OfstJmPCICfg, &val_32);
			*Value = val_32;
			break;
		default:
			return ERROR_INVAILD_PARAMETER;
	}
	return ERROR_NONE;
}

u16 JmWritePciCnfg(struct pci_dev *pdev ,const u32 Value,const u32 OfstJmPCICfg,const u8 length)
{
	u8  val_8;
	u16 val_16;
	u32 val_32;

	val_8 = 0; val_16 = 0; val_32 = 0;
	if (length>4)
		return ERROR_INVAILD_PARAMETER;

	switch (length){
		case 1:
			val_8 = Value;
			pci_write_config_byte(pdev, OfstJmPCICfg, val_8);
			break;
		case 2:
			val_16 = Value;
			pci_write_config_word(pdev, OfstJmPCICfg, val_16);
			break;
		case 3:
		case 4:
			val_32 = Value;
			pci_write_config_dword(pdev, OfstJmPCICfg, val_32);
			break;
		default:
			return ERROR_INVAILD_PARAMETER;
	}
	return ERROR_NONE;
}


void GetChipId(struct pci_dev *pdev ,u32 *pChipId)
{
	u32 dwPage, ChipID;
	u16 status;

	*pChipId = 0;
    status = JmReadPciCnfg(pdev, &dwPage, JMCR_PCICNFG_PAGESEL_OFFSET, 1);
    if (!status)
    {
        // Set Page 1
        dwPage &= ~JMCR_CFGPAGE_MASK;
        dwPage |= (JMCR_CFGPAGE_CHIPID << JMCR_CFGPAGE_SHIFT);
        status = JmWritePciCnfg(pdev, dwPage, JMCR_PCICNFG_PAGESEL_OFFSET, 1);

        // ClkDelay[4] = reg_ec_p7[0]
        status = JmReadPciCnfg(pdev, &ChipID, JMCR_PCICNFG_PAGE_DATA0_OFFSET, 4);
		*pChipId = ChipID;
    }
    else
    {
        UHSDBG ("SetClockDelayValue - Fail to access PCI Config for PageSel!\n");
    }
    // Recover PageSel Setting
    status = JmReadPciCnfg(pdev, &dwPage, JMCR_PCICNFG_PAGESEL_OFFSET, 1);
    dwPage &= ~JMCR_CFGPAGE_MASK;
    status = JmWritePciCnfg(pdev, dwPage, JMCR_PCICNFG_PAGESEL_OFFSET, 1);
}

// Set Timing Delay value by write PCI Config
void SetClockDelayValue(struct pci_dev *pdev,u16 wValue, u8 bDoubleValue,u8 m_SlotIndex)
{
    u32   status;
    u16        wDelay;
	u32			ChipId, dwPage, dwCfgEC, dwCfgB0;


    UHSDBG(" SetClockDelayValue to 0x%02x, DoubleValue = %d\n", wValue, bDoubleValue);

	GetChipId(pdev ,&ChipId);
    if (((ChipId & 0xFFFF0000) == 0x380B0000) ||
        ((ChipId & 0xFFFF0000) == 0x389A0000) ||
        ((ChipId & 0xFFFF0000) == 0x251C0000))
		{

        if (bDoubleValue) // Extend Timing Delay Bits
        {
            wDelay = (wValue & 0x0F) << 1;
        }
        else
        {
            wDelay = (wValue & 0x1F);
        }                                                                                                                  

        status = JmReadPciCnfg(pdev, &dwPage, JMCR_PCICNFG_PAGESEL_OFFSET, 1);
        if (!status)
        {
            // Set Clock Delay Control Bits in Page 7
            dwPage &= ~JMCR_CFGPAGE_MASK;
            dwPage |= (JMCR_CFGPAGE_PAD_DELAY_CTRL2 << JMCR_CFGPAGE_SHIFT);
            status = JmWritePciCnfg(pdev, dwPage, JMCR_PCICNFG_PAGESEL_OFFSET, 1);

            // ClkDelay[4] = reg_ec_p7[0]
            status = JmReadPciCnfg(pdev, &dwCfgEC, JMCR_PCICNFG_PAGE_DATA1_OFFSET, 4);
            if (wDelay & 0x10)
            {
                dwCfgEC |= 0x01;
            }
            else
            {
                dwCfgEC &= ~(0x01);
            }
            status = JmWritePciCnfg(pdev, dwCfgEC, JMCR_PCICNFG_PAGE_DATA1_OFFSET, 4);
        }
        else
        {
            UHSDBG ("SetClockDelayValue - Fail to access PCI Config for PageSel!\n");
        }

        // Recover PageSel Setting
        status = JmReadPciCnfg(pdev, &dwPage, JMCR_PCICNFG_PAGESEL_OFFSET, 1);
        dwPage &= ~JMCR_CFGPAGE_MASK;
        status = JmWritePciCnfg(pdev, dwPage, JMCR_PCICNFG_PAGESEL_OFFSET, 1);

	} // ~ 380B extended Timing Delay Setting
    else
    {
        wDelay = wValue;
    }
		
    //
    //  Set Clock Delay Setting [3:0] in PCI Config B1h
    //
    status = JmReadPciCnfg(pdev, &dwCfgB0, PCICNFG_REG_TIMING_DELAY, 2);
	
    if (status)
    {
        UHSDBG("CHCBase::SetClockDelayValue - ReadPciCnfg Failed!\n");
        return;
    }

    if (m_SlotIndex == 0) // Slot A
    {
        wDelay &= 0x0F;
        wDelay <<= TIMING_DELAY_SLOTA_SHIFT;
        dwCfgB0 &= ~(TIMING_DELAY_BIT_MASK_SLOTA);
    }
    else // Slot B
    {
        wDelay &= 0x0F;
        wDelay <<= TIMING_DELAY_SLOTB_SHIFT;
        dwCfgB0 &= ~(TIMING_DELAY_BIT_MASK_SLOTB);
    }
    dwCfgB0 |= wDelay;

    status = JmWritePciCnfg(pdev, dwCfgB0, PCICNFG_REG_TIMING_DELAY, 2);
    if (status)
    {
        UHSDBG("CHCBase::SetClockDelayValue - WritePciCnfg Failed!\n");
    }
	return;
}

// Set Command Delay Value by write PCI Config. (380B)
void SetCmdDelayValue(struct pci_dev *pdev, u16 wValue)
{
    u32   status;
	u32			dwPage,dwCfgEC ;
    status = JmReadPciCnfg(pdev, &dwPage, JMCR_PCICNFG_PAGESEL_OFFSET, 1);
    if (!status)
    {
        // Set Timing Delay Control Bits in Page 6
        dwPage &= ~JMCR_CFGPAGE_MASK;
        dwPage |= (JMCR_CFGPAGE_PAD_DELAY_CTRL << JMCR_CFGPAGE_SHIFT);
        status = JmWritePciCnfg(pdev, dwPage, JMCR_PCICNFG_PAGESEL_OFFSET, 1);

        // CmdDelay[3:0] = reg_ec_p6[3:0]
        status = JmReadPciCnfg(pdev, &dwCfgEC, JMCR_PCICNFG_PAGE_DATA1_OFFSET, 4);
        dwCfgEC &= ~0x0000000F;
        dwCfgEC |= wValue & 0x0F;
        status = JmWritePciCnfg(pdev, dwCfgEC, JMCR_PCICNFG_PAGE_DATA1_OFFSET, 4);
        // Set Timing Delay Control Bits in Page 7
        dwPage &= ~JMCR_CFGPAGE_MASK;
        dwPage |= (JMCR_CFGPAGE_PAD_DELAY_CTRL2 << JMCR_CFGPAGE_SHIFT);
        status = JmWritePciCnfg(pdev, dwPage, JMCR_PCICNFG_PAGESEL_OFFSET, 1);

        // CmdDelay[4] = reg_ec_p7[1]
        status = JmReadPciCnfg(pdev, &dwCfgEC, JMCR_PCICNFG_PAGE_DATA1_OFFSET, 4);
        if (wValue & 0x10)
        {
            dwCfgEC |= 0x02;
        }
        else
        {
            dwCfgEC &= ~(0x02);
        }
        status = JmWritePciCnfg(pdev, dwCfgEC, JMCR_PCICNFG_PAGE_DATA1_OFFSET, 4);
        // Recover PageSel Setting
        status = JmReadPciCnfg(pdev, &dwPage, JMCR_PCICNFG_PAGESEL_OFFSET, 1);
        dwPage &= ~JMCR_CFGPAGE_MASK;
        status = JmWritePciCnfg(pdev, dwPage, JMCR_PCICNFG_PAGESEL_OFFSET, 1);
    }
    else
    {
        UHSDBG ("SetCmdDelayValue - Fail to access PCI Config for PageSel!\n");
    }
}

// Set Data Delay Value by write PCI Config. (380B)
void SetDataDelayValue(struct pci_dev *pdev,u16 wValue)
{
    u32   	status;
    u32      	dwValue;
	u32			dwPage,dwCfgE8;

    dwValue = wValue & 0x1F;
    status = JmReadPciCnfg(pdev, &dwPage, JMCR_PCICNFG_PAGESEL_OFFSET, 1);
    if (!status)
    {
        // Set Timing Delay Control Bits in Page 6
        dwPage &= ~JMCR_CFGPAGE_MASK;
        dwPage |= (JMCR_CFGPAGE_PAD_DELAY_CTRL << JMCR_CFGPAGE_SHIFT);
        status = JmWritePciCnfg(pdev, dwPage, JMCR_PCICNFG_PAGESEL_OFFSET, 1);

        // DatDelay00 = reg_e8_p6[4:0]
        // DatDelay01 = reg_e8_p6[12:8]
        // DatDelay02 = reg_e8_p6[20:16]
        // DatDelay03 = reg_e8_p6[28:24]
        status = JmReadPciCnfg(pdev, &dwCfgE8, JMCR_PCICNFG_PAGE_DATA0_OFFSET, 4);
        dwCfgE8 &= ~0x1F1F1F1F;
        dwCfgE8 |= dwValue;
        dwCfgE8 |= dwValue << 8;
        dwCfgE8 |= dwValue << 16;
        dwCfgE8 |= dwValue << 24;
        status = JmWritePciCnfg(pdev, dwCfgE8, JMCR_PCICNFG_PAGE_DATA0_OFFSET, 4);
        // Set Timing Delay Control Bits in Page 7
        dwPage &= ~JMCR_CFGPAGE_MASK;
        dwPage |= (JMCR_CFGPAGE_PAD_DELAY_CTRL2 << JMCR_CFGPAGE_SHIFT);
        status = JmWritePciCnfg(pdev, dwPage, JMCR_PCICNFG_PAGESEL_OFFSET, 1);

        // DatDelay08 = reg_e8_p7[4:0]
        // DatDelay09 = reg_e8_p7[12:8]
        // DatDelay10 = reg_e8_p7[20:16]
        // DatDelay11 = reg_e8_p7[28:24]
        status = JmReadPciCnfg(pdev, &dwCfgE8, JMCR_PCICNFG_PAGE_DATA0_OFFSET, 4);
        dwCfgE8 &= ~0x1F1F1F1F;
        dwCfgE8 |= dwValue;
        dwCfgE8 |= dwValue << 8;
        dwCfgE8 |= dwValue << 16;
        dwCfgE8 |= dwValue << 24;
        status = JmWritePciCnfg(pdev, dwCfgE8, JMCR_PCICNFG_PAGE_DATA0_OFFSET, 4);
        // Recover PageSel Setting
        status = JmReadPciCnfg(pdev, &dwPage, JMCR_PCICNFG_PAGESEL_OFFSET, 1);
        dwPage &= ~JMCR_CFGPAGE_MASK;
        status = JmWritePciCnfg(pdev, dwPage, JMCR_PCICNFG_PAGESEL_OFFSET, 1);
    }
    else
    {
        UHSDBG("SetDataDelayValue - Fail to access PCI Config for PageSel!\n");
    }
}

// Set IC Driving
void SetIcDriving(struct pci_dev *pdev,u32 dwDriving)
{
    u32 	status;
    u32   dwVal, dwTmp,ChipId;
	
	if (pdev==NULL){
		printk(KERN_ERR "pdev = NULL \n");
			return;
		}
	GetChipId(pdev,&ChipId);
    if (((ChipId & 0xFFFF0000) == 0x380B0000) ||
        ((ChipId & 0xFFFF0000) == 0x389A0000) ||
        ((ChipId & 0xFFFF0000) == 0x251C0000))
    {

      // [380B] 0xAABBCDEF
      //  AA D[2:0] -> CMD/BS,  PCI Config Offset 0xE4 D[30:28] // Default 3: MAX: 7
      //  BB D[2:0] -> Clock,   PCI Config Offset 0xB8 D[22:20]
      //  C D[2:0] -> Data[0],  PCI Config Offset 0xE4 D[18:16]
      //  D D[2:0] -> Data[1],  PCI Config Offset 0xE4 D[21:19]
      //  E D[2:0] -> Data[2],  PCI Config Offset 0xE4 D[24:22]
      //  F D[2:0] -> Data[3],  PCI Config Offset 0xE4 D[27:25]
      //  X D[2:0] -> Data[4],  PCI Config Offset 0xE? D[?]  // only MMC has 8-bit
      //  X D[2:0] -> Data[5],  PCI Config Offset 0xE? D[?]  // No need to modify Data[7:4] driving for SDXC
      //  X D[2:0] -> Data[6],  PCI Config Offset 0xE? D[?]
      //  X D[2:0] -> Data[7],  PCI Config Offset 0xE? D[?]

      // [Notice] [CTWu@20091024] Joesh suggest to use default Driving if < 100MHz and use MAX Driving if >= 100MHz.
        status = JmReadPciCnfg(pdev, &dwVal, 0xE4, 4);
        if (status)
        {
            UHSDBG("ReadPciCnfg 0xE4 Failed!\n");
            return;
        }

        dwVal &= ~JMCR_IC_DRIVING_E4_MASK_380B;
        dwTmp  = ((dwDriving >> 24) & 0x07) << 28; // CMD/BS
        dwVal |= dwTmp;
        dwTmp  = ((dwDriving >> 12) & 0x07) << 16; // Data[0]
        dwVal |= dwTmp;
        dwTmp  = ((dwDriving >>  8) & 0x07) << 19; // Data[1]
        dwVal |= dwTmp;
        dwTmp  = ((dwDriving >>  4) & 0x07) << 22; // Data[2]
        dwVal |= dwTmp;
        dwTmp  = (dwDriving & 0x07) << 25;         // Data[3]
        dwVal |= dwTmp;

        UHSDBG("SetIcDriving_380B - Set E4h to 0x%08x\n", dwVal);
        status = JmWritePciCnfg(pdev, dwVal, 0xE4, 4);

        if (status)
        {
            UHSDBG("WritePciCnfg 0xE4 Failed!\n");
            return;
        }
        status = JmReadPciCnfg(pdev, &dwVal, 0xB8, 4);
        if (status)
        {
            UHSDBG("ReadPciCnfg 0xB8 Failed!\n");
            return;
        }

        dwVal &= ~JMCR_IC_DRIVING_B8_MASK_380B;
        dwTmp  = ((dwDriving >> 16) & 0x07) << 20; // Clock
        dwVal |= dwTmp;

        UHSDBG("SetIcDriving_380B - Set B8h to 0x%08x\n", dwVal);
        status = JmWritePciCnfg(pdev, dwVal, 0xB8, 4);
        if (status)
        {
            UHSDBG("WritePciCnfg 0xB8 Failed!\n");
            return;
        }
    }
    else // 385A/B, 380A
    {
      //  DWord  0xAABBCCDD
      //  AA D[1:0] -> CMD/BS,    PCI Config Offset 0xE4 D[3:2]
      //  BB D[1:0] -> Clock,     PCI Config Offset 0xB8 D[23:22]
      //  CC D[1:0] -> Data[3:0], PCI Config Offset 0xE4 D[1:0]
      //  DD D[1:0] -> Data[7:4], PCI Config Offset 0xE4 D[5:4]
        status = JmReadPciCnfg(pdev, &dwVal, 0xE4, 4);
        if (status)
        {
            UHSDBG("ReadPciCnfg 0xE4 Failed!\n");
            return;
        }
        dwVal &= ~JMCR_IC_DRIVING_E4_MASK;
        dwTmp  = ((dwDriving >> 24) & 0x03) << 2; // CMD/BS
        dwVal |= dwTmp;
        dwTmp  = (dwDriving >> 8) & 0x03;       // Data[3:0]
        dwVal |= dwTmp;
        dwTmp  = (dwDriving & 0x03) << 4;       // Data[7:4]
        dwVal |= dwTmp;
        status = JmWritePciCnfg(pdev, dwVal, 0xE4, 4);
        if (status)
        {
            UHSDBG("WritePciCnfg 0xE4 Failed!\n");
            return;
        }

        status = JmReadPciCnfg(pdev, &dwVal, 0xB8, 4);
        if (status)
        {
            UHSDBG ("ReadPciCnfg 0xB8 Failed!\n");
            return;
        }
        dwVal &= ~JMCR_IC_DRIVING_B8_MASK;
        dwTmp  = ((dwDriving >> 16) & 0x03) << 22; // Clock
        dwVal |= dwTmp;
        status = JmWritePciCnfg(pdev, dwVal, 0xB8, 4);
        if (status)
        {
            UHSDBG("WritePciCnfg 0xB8 Failed!\n");
            return;
        }
    }
}
EXPORT_SYMBOL(SetIcDriving);

void SetTimeoutTime(struct sdhci_host *sdhost, u16 valTimeOut)
{
		sdhci_writeb(sdhost, 0xE, SDHCI_TIMEOUT_CONTROL);
}

void SetExternOE(struct sdhci_host *sdhost, u8 val)
{
	u32 tmpVal;

	tmpVal = sdhci_readl(sdhost ,  JMADDR_EXTERN_OE );	
	if (val){	
		tmpVal |= 1<<23;
		tmpVal |= 1<<22;	
	}else{
		tmpVal &= ~(1<<23);
		tmpVal &= ~(1<<22);		
	}
	sdhci_writel(sdhost , tmpVal , JMADDR_EXTERN_OE );
}
void GlobalSetClockDelayValue(struct mmc_host *mmchost, u16 wValue)
{
   	struct pci_dev *pdev = NULL;
	struct sdhci_pci_slot *slot = NULL;
			

		/* if we have last tuning result	*/
		slot = sdhci_priv(mmc_priv(mmchost));
		if (slot==NULL){
			pr_debug("%s:%s():sdhci host error! Thers is no slot!! \n",__FILE__,__FUNCTION__);
			return;
		}
		pdev = slot->chip->pdev;	
		 SetClockDelayValue( pdev,wValue, true,0);
		 SetCmdDelayValue( pdev, 0);
		 SetDataDelayValue( pdev, 0);

}
EXPORT_SYMBOL(GlobalSetClockDelayValue);


u32	SendCommand(struct sdhci_host *sdhost, u8 opcode,u32 typeResponse,u32 *response)
{
	int err;
	struct mmc_command cmd;

	memset(&cmd, 0, sizeof(struct mmc_command));
	cmd.opcode = opcode;
	cmd.arg = 0;
	cmd.flags = typeResponse;
	if (opcode == MMC_SEND_STATUS){
		cmd.flags |= MMC_CMD_AC;
		cmd.arg    = sdhost->mmc->card->rca << 16;
	}else if (opcode == MMC_STOP_TRANSMISSION)
		cmd.flags |= MMC_CMD_AC;

	mmc_claim_host(sdhost->mmc);
	err = mmc_wait_for_cmd(sdhost->mmc, &cmd, MMC_CMD_RETRIES);
	mmc_release_host(sdhost->mmc);
	if (err)
		return err;
	memcpy(response , cmd.resp,4);
	return ERROR_NONE;
}
EXPORT_SYMBOL(SendCommand);

extern int mmc_sd_resume(struct mmc_host *host);

void ResetHost(struct sdhci_host *sdhost)
{
		u32  clock;
		
		UHSDBG("Enter ResetHost() \n");
		clock = sdhost->mmc->ios.clock;
		mmc_power_off(sdhost->mmc);
		msleep(50);
		mmc_power_up(sdhost->mmc);
		msleep(50);
		sdhost->mmc->ops->reinit(sdhost->mmc); 	
		sdhci_writeb(sdhost, SDHCI_POWER_330|SDHCI_POWER_ON, SDHCI_POWER_CONTROL);
		msleep(600);
		mmc_sd_resume(sdhost->mmc);
		sdhost->mmc->tuning_state = 1;
		msleep(50);
		mmc_set_clock(sdhost->mmc, clock);			/*	Default 100MHZ or 25MHZ after initiation, set clock as original value */
		return ;
}
EXPORT_SYMBOL(ResetHost);


#define BUFFER_SIZE				(JM_NUM_BLK_MULTI_READ*JM_BLOCK_SIZE)		/*128 block == 64K */
int ReadData(struct sdhci_host *sdhost,u8 *buf_DMA)
{	
	struct mmc_request mrq;
	struct mmc_command cmd;
	struct mmc_command stop;
	struct mmc_data data;
	struct scatterlist sg;
	int 	retVal,count_loop = 0;
	u32 	response[4];
	u8	*buffer;	
	
	retVal = ERROR_NONE;
	response[0] = 0;response[1] = 0;response[2] = 0;response[3] = 0;
	memset(buf_DMA,0,BUFFER_SIZE);
	
	buffer = kzalloc(JM_NUM_BLK_MULTI_READ*JM_BLOCK_SIZE, GFP_KERNEL);	
	if (!buffer){
		return ERROR_NO_MEMORY;
	}
	mmc_claim_host(sdhost->mmc);
	
	memset(&cmd, 0, sizeof(struct mmc_command));
	cmd.opcode = MMC_SET_BLOCKLEN;
	cmd.arg = JM_BLOCK_SIZE;
	cmd.flags = MMC_RSP_R1 | MMC_CMD_AC;
	retVal = mmc_wait_for_cmd(sdhost->mmc, &cmd, 1);
	if (retVal)
		goto exit_readdata;
	sg_init_one(&sg, buffer, JM_NUM_BLK_MULTI_READ*JM_BLOCK_SIZE);

	/*  mmc_test_simple_transfer(test, &sg, 1, 0, 3, 512, 0);        test->card   */
	memset(&mrq, 0, sizeof(struct mmc_request));
	memset(&cmd, 0, sizeof(struct mmc_command));
	memset(&data, 0, sizeof(struct mmc_data));
	memset(&stop, 0, sizeof(struct mmc_command));
	mrq.cmd = &cmd;
	mrq.data = &data;
	mrq.stop = &stop;

/*
sg_len	= 1
dev_addr= 0
blocks	= JM_NUM_BLK_MULTI_READ = 3
blksz 	=JM_BLOCK_SIZE = 512
write 	= 0
*/


	mrq.cmd->opcode = MMC_READ_MULTIPLE_BLOCK;
	mrq.cmd->arg = 0;

	mrq.cmd->flags = MMC_RSP_R1 | MMC_CMD_ADTC;
	mrq.stop->opcode = MMC_STOP_TRANSMISSION;
	mrq.stop->arg = 0;
	mrq.stop->flags = MMC_RSP_R1B | MMC_CMD_AC;
		

	mrq.data->blksz = JM_BLOCK_SIZE;
	mrq.data->blocks = JM_NUM_BLK_MULTI_READ;
	mrq.data->flags = MMC_DATA_READ;
	mrq.data->sg = &sg;
	mrq.data->sg_len = 1;
	mmc_set_data_timeout(mrq.data, sdhost->mmc->card);
	mmc_wait_for_req(sdhost->mmc, &mrq);	
	/*mmc_test_wait_busy(test);*/
	do {
		SendCommand(sdhost, MMC_SEND_STATUS, MMC_RSP_R1, response);
		mdelay(5);
	} while ((!(response[0] & R1_READY_FOR_DATA)) && (++count_loop<10));
	if	(!(count_loop<10)){
			retVal = ERROR_MMC_COMMAND_TIMEOUT;
			goto exit_readdata;
	}
	/* mmc_test_check_result(test, &mrq);*/
	retVal = 0;
	if (!retVal && mrq.cmd->error)
		retVal = OFFSET_CMD_ERROR + mrq.cmd->error;
	if (!retVal && mrq.data->error)
		retVal = OFFSET_DATA_ERROR + mrq.data->error;
	if (!retVal && mrq.stop && mrq.stop->error)
		retVal = mrq.stop->error;
	if (!retVal && mrq.data->bytes_xfered != mrq.data->blocks * mrq.data->blksz)
		retVal = ERROR_XFERSIZE_MISMACH;

	if (retVal == -EINVAL)
		goto exit_readdata;
		
	sg_copy_to_buffer(&sg, 1, buf_DMA, BUFFER_SIZE);
	
exit_readdata:
	kfree(buffer);
	mmc_release_host(sdhost->mmc);
	return retVal;
}

u8 CheckCardPresent(struct sdhci_host *sdhost)
{	
	return (sdhci_readl(sdhost, SDHCI_PRESENT_STATE)& SDHCI_CARD_PRESENT) ? true:false;
}
EXPORT_SYMBOL(CheckCardPresent);

/*	if response type == R1b , return [39:8], but we don't need [39:32]	*/
void SDGetCardStatusFromResponse(u32 *response, u16 *cardStatus)
{
	*cardStatus = (response[0] & 0x00FFFF00)>>8;
}

struct tuning_setting lastResult;
void SaveAdjustResult(struct sdhci_host *sdhost, const u8 bClkDelay, const u8 bCmdDelay, const u32 clock)
{
	UHSDBG("%s: ClkDelay=%x , CmdDelay=%x , DataDelay=%x , Clock=%x \n",__FUNCTION__, bClkDelay,
																														bCmdDelay ,bCmdDelay ,clock );

		if (clock){
		lastResult.clock 						= clock;
		if (bClkDelay!=((u8)-1))		
			lastResult.delay_clock 		= bClkDelay;
		if (bCmdDelay!=((u8)-1))
			lastResult.delay_command = bCmdDelay;
		if (bCmdDelay!=((u8)-1))
			lastResult.delay_data 			= bCmdDelay;
	}
	return ;
}

u8 setLastTuningResult(struct sdhci_host *sdhost)
{
	struct pci_dev *pdev = NULL;
	struct sdhci_pci_slot *slot = NULL;
	
	UHSDBG("%s: clock = %d  clkdelay=%d delay_data=%d\n",__FUNCTION__, lastResult.clock, lastResult.delay_clock, lastResult.delay_data);				
	if (lastResult.clock > 0){
		/* if we have last tuning result	*/
		slot = sdhci_priv(sdhost);
		if (slot==NULL){
			pr_debug("%s:%s():sdhci host error! Thers is no slot!! \n",__FILE__,__FUNCTION__);
			return ERROR_INVAILD_PARAMETER;
		}
		pdev = slot->chip->pdev;	

		mmc_set_clock(sdhost->mmc, lastResult.clock);
		if (sdhost->mmc->ios.clock >= JM_CLOCK_100MHZ){
			/* set last result*/
			if (lastResult.delay_clock!=((u8)-1))
				SetClockDelayValue( pdev,lastResult.delay_clock, false,0);		
			if (lastResult.delay_command!=((u8)-1))
				SetCmdDelayValue(pdev  ,lastResult.delay_command );
			if (lastResult.delay_data!=((u8)-1))
				SetDataDelayValue(pdev ,lastResult.delay_data );
		}else{
			if (lastResult.delay_clock!=((u8)-1))
				SetClockDelayValue( pdev,lastResult.delay_clock, true,0);		
		}
		return ERROR_NONE;
	}else
		return ERROR_INVAILD_SETTING;
}

u32	AdjustTimingDelay(struct sdhci_host *sdhost,u32 clock)
{
	u16 			wPhaseCount;
    u8            bValue, CurOkStart, CurContOkCount, MaxContOkStart, MaxContOkCount;
    u8            bClkDelay , bCmdDelay, i;
    u8         fPrevOK, bStopAdjust=false;
    u8         bUseCmdDelay, bDoubleValue = true;
    u32           dwTestResult, dwTestBit ,dwTestResult_Cmd;
    u16      		cardStatus;
    u32 			response[4] ,status;
	u8 				*pVerifyData,*pDmaBuffer;
	u32				retVal,retVal_fun;
	struct sdhci_pci_slot *slot = NULL;
	struct pci_dev *pdev = NULL;

	pVerifyData = kzalloc(BUFFER_SIZE, GFP_KERNEL);
	pDmaBuffer = kzalloc(BUFFER_SIZE, GFP_KERNEL);
	bClkDelay = (u8)-1;
	bCmdDelay = (u8)-1;
	if (sdhost->mmc->card==NULL){
		pr_debug("%s:invaild card, \n",__FUNCTION__);
		retVal = ERROR_INVAILD_CARD;
		goto err_exit;
	}
    if (clock >= JM_CLOCK_100MHZ) // For 380B/380A Chips
    {
        wPhaseCount  = 32;
        bDoubleValue = false;
        bUseCmdDelay = true;
    }
    else // For Compatible with 385A/385B/380A
    {
        wPhaseCount  = 16;
        bDoubleValue = true;
        bUseCmdDelay = false;
    }

	slot = sdhci_priv(sdhost);
	if (slot==NULL){
		printk(KERN_ERR "%s:%s():sdhci host error! Thers is no slot!! \n",__FILE__, __FUNCTION__);
		retVal = ERROR_NO_SLOT;
		goto err_exit;
	}
	pdev = slot->chip->pdev;
    // Set shorter Command Timeout time to reduce time needed with Adjust Timing Delay
	SetTimeoutTime(sdhost, 500);
    mmc_set_clock(sdhost->mmc, clock);	

	if (clock<=JM_CLOCK_25MHZ){
		retVal = ERROR_NONE;
		goto err_exit;
	}
				
    dwTestResult = 0;
    dwTestBit    = 1;
    dwTestResult_Cmd = 0;
    for(i=0 ; i<wPhaseCount ; i++, dwTestBit <<= 1) // Timing Delay Setting value is from 0x00 to 0x1F
    {
        if (CheckCardPresent(sdhost) == false)
        {
            pr_debug ("Found Card Removed during AdjustTiming!\n");
            dwTestResult = 0;
			retVal = ERROR_CARDREMOVE;
			goto err_exit;
        }
        SetClockDelayValue( pdev,i, bDoubleValue,0);
        if (bUseCmdDelay)
        {
            SetCmdDelayValue(pdev ,0);
            SetDataDelayValue(pdev ,0);
        }
        sdhost->mmc->ops->soft_reset(sdhost->mmc,SDHCI_RESET_CMD|SDHCI_RESET_DATA );		
        status = SendCommand(sdhost, MMC_SEND_STATUS, MMC_RSP_R1, response);
        if (!status)
        {
            SDGetCardStatusFromResponse(response, &cardStatus);

            // Issue Stop CMD if at Data Transfer State
            if (((cardStatus & SD_STATUS_CURRENT_STATE_MASK) == 0x0A00) ||
                ((cardStatus & SD_STATUS_CURRENT_STATE_MASK) == 0x0C00))
            {
                status = SendCommand(sdhost, MMC_STOP_TRANSMISSION, MMC_RSP_R1B, response);
                status = SendCommand(sdhost, MMC_SEND_STATUS, MMC_RSP_R1, response);
                pr_debug("Got Sd CardStatus= 0x%08x after CMD12\n", cardStatus);
                //if (cardStatus != 0x00000900) // SD_STATUS_READY_FOR_DATA | (SD_STATUS_CURRENT_STATE_TRAN << 9)
                if (((cardStatus & SD_STATUS_CURRENT_STATE_MASK) == 0x0A00) ||
                    ((cardStatus & SD_STATUS_CURRENT_STATE_MASK) == 0x0C00))
                {
                    pr_debug("ResetHost for SdStaus Error after CMD12!\n");
                    ResetHost(sdhost);
                    continue;
                }
            }
        }
        else
        {
            pr_debug("Send SD SendStatus Cmd Failed! (%d)\n", i);
            continue;
        }

        status = ReadData(sdhost,pDmaBuffer);
        if (!status){
            // Read Twice for Safe
            status = ReadData(sdhost,pVerifyData);
            if (!status){
                retVal_fun = memcmp(pVerifyData, pDmaBuffer, BUFFER_SIZE);
                if (retVal_fun!=FUNRT_MEMCMP_SAME)
                {
                    pr_debug("Data Verified Error!\n");
                    status = ERROR_DATA_MISMATCH;
                }
            }
        }
        if (!status) // Set result bit of current setting value as Success
        {
            dwTestResult |= dwTestBit;
            dwTestResult_Cmd |= dwTestBit;
        }
        else if ((status > OFFSET_CMD_ERROR) && (status < OFFSET_DATA_ERROR))
        {
            // Set CMD Result as Success when only DATA Error.
            dwTestResult_Cmd |= dwTestBit;
        }
        else if (status == ERROR_DATA_MISMATCH)
        {
            pr_debug ("Stop Adjust for Data Compare Error!\n");
            bStopAdjust = true;
            break;
        }
        /*else if (status == JM_STATUS_TIMER_TIMEOUT)
        {
            SdDBGOut(DBG_WARN, ("ResetHost and Stop Adjust for Request Timeout!\n"));
            ResetHost(TRUE);
            //bStopAdjust = TRUE;
            //break;
        }*/

      // For DOS, always ResetHost with Error Case for more accuracy result?
      //#ifndef WINDOWS_DRV
      //  if (!JM_SUCCESS(status))
      //  {
      //      ResetHost(TRUE);
      //  }
      //#endif // WINDOWS_DRV
    } //~ for(i=0 ; i<wPhaseCount ; i++, dwTestBit <<= 1)	
	
    SetTimeoutTime(sdhost ,JMCR_CMD_TIMEOUT_TIME);  // Restore Default Command Timeout Time.
    if (bStopAdjust)
    {
        ResetHost(sdhost);
		retVal = ERROR_STOP_ADJUSTMENT;
		goto err_exit;
    }
	
    CurOkStart     = 0;
    CurContOkCount = 0;
    MaxContOkStart = 0;
    MaxContOkCount = 0;
    // [Notice] wPhaseCount is always 32
    // Find Range of Successful Timing Delay Setting
		
    if (dwTestResult == 0) // All Failed!
    {
        MaxContOkCount = 0;
    }
    else if (dwTestResult == 0xFFFFFFFF)
    {
        MaxContOkCount = 31;
    }
    else // Find Largest Range of Successful Timing Delay Setting
    {
        fPrevOK   = false;
        for(bValue=0, dwTestBit=1 ; bValue<wPhaseCount ; bValue++, dwTestBit<<=1)
        {
            if (dwTestResult & dwTestBit) // Result Current Setting Value is Successed
            {
                if (fPrevOK == false)
                {
                    CurOkStart     = bValue;
                    CurContOkCount = 1;
                    fPrevOK        = true;
                }
                else
                {
                    CurContOkCount++;
                    if (CurContOkCount > MaxContOkCount)
                    {
                        MaxContOkStart = CurOkStart;
                        MaxContOkCount = CurContOkCount;
                    }
                }
            }
            else // Result Current Setting Value is Failed
            {
                fPrevOK = false;
            }
        }
    }
	ResetHost(sdhost);
	mmc_set_clock(sdhost->mmc, clock);
	UHSDBG("Clock Test Result: 0x%08x,  Cmd Test Result: 0x%08x ,wPhaseCount=%d \n", dwTestResult, dwTestResult_Cmd,wPhaseCount);
    // Select Best Setting from Succeed Setting Range
    if (MaxContOkCount == 0)
    {
        pr_debug ("Cannot Find Succeed Clock Delay Setting!?\n");
		retVal = ERROR_NO_CLOCK_DELAY_SETTING;
		goto err_exit;
    }
    bClkDelay = MaxContOkStart + ((MaxContOkCount-1) / 2);
    UHSDBG("Clock Delay Success Range %02xh to %02xh. Select %02xh.\n",
                        MaxContOkStart, MaxContOkStart + MaxContOkCount - 1, bClkDelay);
					
    // Set Cmd and Data Delay with Adjusted Value
    UHSDBG ("Set ClkDelay= %02xh at %d Hz\n", bClkDelay, clock);
    SetClockDelayValue(pdev, bClkDelay, bDoubleValue,0);
    // Calculate CMD Delay and Data Delay for 380B >= 83MHz case
    if (bUseCmdDelay)
    {
        if (dwTestResult_Cmd == 0) // All Failed!
        {
            MaxContOkCount = 0;
        }
        else if (dwTestResult_Cmd == 0xFFFFFFFF)
        {
            MaxContOkCount = 31;
        }
        else // Find Cmd Timing Range by Successful Cmd Delay Result + Selected Clock Range
        {
            if (MaxContOkStart > 0) // Find Cmd's MaxContOkStart by Clk's MaxContOkStart
            {
                CurOkStart = MaxContOkStart -1;
                dwTestBit = 1 << CurOkStart;
                for(bValue=CurOkStart ; bValue>0 ; bValue--, dwTestBit>>=1)
                {
                    if ((dwTestResult_Cmd & dwTestBit) == 0)
                    {
                        break;
                    }
                }
                dwTestBit = 1 << bValue;
                if ((dwTestResult_Cmd & dwTestBit) == 0)
                {
                    bValue++;
                }
                CurOkStart = bValue;
            }
			else
			{
			    CurOkStart = 0;
			}

            bValue    = MaxContOkStart + MaxContOkCount;
            dwTestBit = 1 << bValue;
            for(; bValue<32 ; bValue++, dwTestBit<<=1)
            {
                if ((dwTestResult_Cmd & dwTestBit) == 0)
                {
                    break;
                }
            }
            MaxContOkStart = CurOkStart;
            MaxContOkCount = bValue - MaxContOkStart;
        } // ~ dwTestResult_Cmd != 0 && != 0xFFFFFFFF

        if (MaxContOkCount == 0)
        {
            pr_debug ("Cannot Find Succeed Cmd Delay Setting!?\n");
            bCmdDelay = 0;
        }
        else
        {
            bCmdDelay = MaxContOkStart + ((MaxContOkCount-1) / 2);
            pr_debug ("Cmd Delay Success Range %02xh to %02xh. Select %02xh.\n",
                                MaxContOkStart, MaxContOkStart + MaxContOkCount - 1, bCmdDelay);

            if (bCmdDelay > bClkDelay)
            {
                bCmdDelay -= bClkDelay;
            }
            else
            {
                bCmdDelay = 0;
            }
            pr_debug ("Modified Cmd Delay = %02xh.\n", bCmdDelay);
        }
        // Set Cmd and Data Delay with Adjusted Value
        SetCmdDelayValue(pdev, bCmdDelay);
        SetDataDelayValue(pdev, bCmdDelay);

    } //~ if (bUseCmdDelay)
	
	SaveAdjustResult(sdhost, bClkDelay , bCmdDelay,clock );
	status = SendCommand(sdhost, MMC_SEND_STATUS, MMC_RSP_R1, response);
    if (status){
		retVal = ERROR_INVAILD_SETTING;
		goto err_exit;
    }else{
		status = ReadData(sdhost,pDmaBuffer);
		if (status){
			retVal = ERROR_INVAILD_SETTING;
			goto err_exit;			
		}
	}
	retVal = ERROR_NONE;

err_exit:
	kfree(pDmaBuffer);
	kfree(pVerifyData);
	return retVal;
}
EXPORT_SYMBOL(AdjustTimingDelay);


/*=================================================*/


void mmc_rescan(struct work_struct *work)
{
	struct mmc_host *host =
		container_of(work, struct mmc_host, detect.work);
	u32 ocr;
	int err;

	mmc_bus_get(host);

	host->tuning_state = 0;
	
	/* if there is a card registered, check whether it is still present */
	if ((host->bus_ops != NULL) && host->bus_ops->detect && !host->bus_dead)
		host->bus_ops->detect(host);

	mmc_bus_put(host);


	mmc_bus_get(host);

	/* if there still is a card present, stop here */
	if (host->bus_ops != NULL) {
		mmc_bus_put(host);
		goto out;
	}

	/* detect a newly inserted card */

	/*
	 * Only we can add a new handler, so it's safe to
	 * release the lock here.
	 */
	mmc_bus_put(host);

	if (host->ops->get_cd && host->ops->get_cd(host) == 0)
		goto out;

	mmc_claim_host(host);

	mmc_power_up(host);
	sdio_reset(host);
	mmc_go_idle(host);

	mmc_send_if_cond(host, host->ocr_avail);

	/*
	 * First we search for SDIO...
	 */
	err = mmc_send_io_op_cond(host, 0, &ocr);
	if (!err) {
		if (mmc_attach_sdio(host, ocr)) {
			mmc_claim_host(host);
			/* try SDMEM (but not MMC) even if SDIO is broken */
			if (mmc_send_app_op_cond(host, 0, &ocr))
				goto out_fail;

			if (mmc_attach_sd(host, ocr))
				mmc_power_off(host);
		}
		goto out;
	}

	/*
	 * ...then normal SD...
	 */
	err = mmc_send_app_op_cond(host, 0, &ocr);
	if (!err) {
		if (mmc_attach_sd(host, ocr))
			mmc_power_off(host);
		goto out;
	}

	/*
	 * ...and finally MMC.
	 */
	err = mmc_send_op_cond(host, 0, &ocr);
	if (!err) {
		if (mmc_attach_mmc(host, ocr))
			mmc_power_off(host);
		goto out;
	}

out_fail:
	mmc_release_host(host);
	mmc_power_off(host);

out:
	if (host->caps & MMC_CAP_NEEDS_POLL)
		mmc_schedule_delayed_work(&host->detect, HZ);
}

void mmc_start_host(struct mmc_host *host)
{
	mmc_power_off(host);
	mmc_detect_change(host, 0);
}

void mmc_stop_host(struct mmc_host *host)
{
#ifdef CONFIG_MMC_DEBUG
	unsigned long flags;
	spin_lock_irqsave(&host->lock, flags);
	host->removed = 1;
	spin_unlock_irqrestore(&host->lock, flags);
#endif

	if (host->caps & MMC_CAP_DISABLE)
		cancel_delayed_work(&host->disable);
	cancel_delayed_work(&host->detect);
	mmc_flush_scheduled_work();

	/* clear pm flags now and let card drivers set them as needed */
	host->pm_flags = 0;

	mmc_bus_get(host);
	if (host->bus_ops && !host->bus_dead) {
		if (host->bus_ops->remove)
			host->bus_ops->remove(host);

		mmc_claim_host(host);
		mmc_detach_bus(host);
		mmc_release_host(host);
		mmc_bus_put(host);
		return;
	}
	mmc_bus_put(host);

	BUG_ON(host->card);

	mmc_power_off(host);
}

void mmc_power_save_host(struct mmc_host *host)
{
	mmc_bus_get(host);

	if (!host->bus_ops || host->bus_dead || !host->bus_ops->power_restore) {
		mmc_bus_put(host);
		return;
	}

	if (host->bus_ops->power_save)
		host->bus_ops->power_save(host);

	mmc_bus_put(host);

	mmc_power_off(host);
}
EXPORT_SYMBOL(mmc_power_save_host);

void mmc_power_restore_host(struct mmc_host *host)
{
	mmc_bus_get(host);

	if (!host->bus_ops || host->bus_dead || !host->bus_ops->power_restore) {
		mmc_bus_put(host);
		return;
	}

	mmc_power_up(host);
	host->bus_ops->power_restore(host);

	mmc_bus_put(host);
}
EXPORT_SYMBOL(mmc_power_restore_host);

int mmc_card_awake(struct mmc_host *host)
{
	int err = -ENOSYS;

	mmc_bus_get(host);

	if (host->bus_ops && !host->bus_dead && host->bus_ops->awake)
		err = host->bus_ops->awake(host);

	mmc_bus_put(host);

	return err;
}
EXPORT_SYMBOL(mmc_card_awake);

int mmc_card_sleep(struct mmc_host *host)
{
	int err = -ENOSYS;

	mmc_bus_get(host);

	if (host->bus_ops && !host->bus_dead && host->bus_ops->awake)
		err = host->bus_ops->sleep(host);

	mmc_bus_put(host);

	return err;
}
EXPORT_SYMBOL(mmc_card_sleep);

int mmc_card_can_sleep(struct mmc_host *host)
{
	struct mmc_card *card = host->card;

	if (card && mmc_card_mmc(card) && card->ext_csd.rev >= 3)
		return 1;
	return 0;
}
EXPORT_SYMBOL(mmc_card_can_sleep);

#ifdef CONFIG_PM

/**
 *	mmc_suspend_host - suspend a host
 *	@host: mmc host
 */
int mmc_suspend_host(struct mmc_host *host)
{
	int err = 0;

	if (host->caps & MMC_CAP_DISABLE)
		cancel_delayed_work(&host->disable);
	cancel_delayed_work(&host->detect);
	mmc_flush_scheduled_work();

	mmc_bus_get(host);
	if (host->bus_ops && !host->bus_dead) {
		if (host->bus_ops->suspend)
			err = host->bus_ops->suspend(host);
		if (err == -ENOSYS || !host->bus_ops->resume) {
			/*
			 * We simply "remove" the card in this case.
			 * It will be redetected on resume.
			 */
			if (host->bus_ops->remove)
				host->bus_ops->remove(host);
			mmc_claim_host(host);
			mmc_detach_bus(host);
			mmc_release_host(host);
			host->pm_flags = 0;
			err = 0;
		}
	}
	mmc_bus_put(host);

	if (!err && !(host->pm_flags & MMC_PM_KEEP_POWER))
		mmc_power_off(host);

	return err;
}

EXPORT_SYMBOL(mmc_suspend_host);

/**
 *	mmc_resume_host - resume a previously suspended host
 *	@host: mmc host
 */
int mmc_resume_host(struct mmc_host *host)
{
	int err = 0;

	mmc_bus_get(host);
	if (host->bus_ops && !host->bus_dead) {
		if (!(host->pm_flags & MMC_PM_KEEP_POWER)) {
			mmc_power_up(host);
			mmc_select_voltage(host, host->ocr);
		}
		BUG_ON(!host->bus_ops->resume);
		err = host->bus_ops->resume(host);
		if (err) {
			printk(KERN_WARNING "%s: error %d during resume "
					    "(card was removed?)\n",
					    mmc_hostname(host), err);
			err = 0;
		}
	}
	mmc_bus_put(host);

	return err;
}
EXPORT_SYMBOL(mmc_resume_host);

/* Do the card removal on suspend if card is assumed removeable
 * Do that in pm notifier while userspace isn't yet frozen, so we will be able
   to sync the card.
*/
int mmc_pm_notify(struct notifier_block *notify_block,
					unsigned long mode, void *unused)
{
	struct mmc_host *host = container_of(
		notify_block, struct mmc_host, pm_notify);
	unsigned long flags;


	switch (mode) {
	case PM_HIBERNATION_PREPARE:
	case PM_SUSPEND_PREPARE:

		spin_lock_irqsave(&host->lock, flags);
		host->rescan_disable = 1;
		spin_unlock_irqrestore(&host->lock, flags);
		cancel_delayed_work_sync(&host->detect);

		if (!host->bus_ops || host->bus_ops->suspend)
			break;

		mmc_claim_host(host);

		if (host->bus_ops->remove)
			host->bus_ops->remove(host);

		mmc_detach_bus(host);
		mmc_release_host(host);
		host->pm_flags = 0;
		break;

	case PM_POST_SUSPEND:
	case PM_POST_HIBERNATION:

		spin_lock_irqsave(&host->lock, flags);
		host->rescan_disable = 0;
		spin_unlock_irqrestore(&host->lock, flags);
		mmc_detect_change(host, 0);

	}

	return 0;
}
#endif

static int __init mmc_init(void)
{
	int ret;

	workqueue = create_singlethread_workqueue("kmmcd");
	if (!workqueue)
		return -ENOMEM;

	ret = mmc_register_bus();
	if (ret)
		goto destroy_workqueue;

	ret = mmc_register_host_class();
	if (ret)
		goto unregister_bus;

	ret = sdio_register_bus();
	if (ret)
		goto unregister_host_class;

	return 0;

unregister_host_class:
	mmc_unregister_host_class();
unregister_bus:
	mmc_unregister_bus();
destroy_workqueue:
	destroy_workqueue(workqueue);

	return ret;
}

static void __exit mmc_exit(void)
{
	sdio_unregister_bus();
	mmc_unregister_host_class();
	mmc_unregister_bus();
	destroy_workqueue(workqueue);
}

subsys_initcall(mmc_init);
module_exit(mmc_exit);

MODULE_LICENSE("GPL");
